﻿namespace TelpoPush.Fence.Worker.Models.Fence
{
    public static class CirclePolygonAlgorithm
    {
        /// <summary>
        /// 判断圆和多边形是否相交
        /// </summary>
        /// <param name="circle">圆心坐标</param>
        /// <param name="radius">圆半径</param>
        /// <param name="polygon">多边形各个点坐标</param>
        /// <returns></returns>
        public static bool IsIntersect(GpsFencePoint circle, double degree, IList<GpsFencePoint> polygon)
        {
            double powRadius = Math.Pow(degree, 2);

            //情况1：多边形任一顶点在圆内，则认为相交
            foreach (var pt in polygon)
            {
                if (Math.Pow(circle.Longitude - pt.Longitude, 2) + Math.Pow(circle.Latitude - pt.Latitude, 2) < powRadius) return true;
            }

            //情况2：多边形顶点集的任一有效线段和圆相交（圆心到线段的距离小于半径）
            var lines = EnumeratePolygonLines(polygon);
            foreach (var seg in lines)
            {
                if (PointToLine(circle, seg.Item1, seg.Item2) < degree) return true;
            }

            return false;
        }

        private static IList<Tuple<GpsFencePoint, GpsFencePoint>> EnumeratePolygonLines(IList<GpsFencePoint> polygon)
        {
            var list = new List<Tuple<GpsFencePoint, GpsFencePoint>>();

            for (int i = 0; i < polygon.Count - 1; i++)
            {
                list.Add(new Tuple<GpsFencePoint, GpsFencePoint>(polygon[i], polygon[i + 1]));
            }
            list.Add(new Tuple<GpsFencePoint, GpsFencePoint>(polygon[polygon.Count - 1], polygon[0]));

            return list;
        }

        //计算两点之间的距离
        private static double LineSpace(GpsFencePoint ptStart, GpsFencePoint ptEnd)
        {
            if (ptStart == ptEnd) return 0;
            return Math.Sqrt((ptEnd.Longitude - ptStart.Longitude) * (ptEnd.Longitude - ptStart.Longitude) + (ptEnd.Latitude - ptStart.Latitude) * (ptEnd.Latitude - ptStart.Latitude));
        }

        /// <summary>
        /// 点到线段的距离
        /// </summary>
        /// <param name="pt"></param>
        /// <param name="ptStart">线段顶点1</param>
        /// <param name="ptEnd">线段顶点2</param>
        /// <returns></returns>
        private static double PointToLine(GpsFencePoint pt, GpsFencePoint ptStart, GpsFencePoint ptEnd)
        {
            double distance = 0;
            double a, b, c;
            a = LineSpace(ptStart, ptEnd);  //线段长度
            b = LineSpace(pt, ptStart);     //点到线段顶点1的距离
            c = LineSpace(pt, ptEnd);       //点到线段顶点2的距离
            if (b == 0 || c == 0)
            {
                distance = 0;
                return distance;
            }
            if (a == 0)
            {
                distance = b;
                return distance;
            }
            if (c * c >= a * a + b * b)
            {
                distance = b;
                return distance;
            }
            if (b * b >= a * a + c * c)
            {
                distance = c;
                return distance;
            }

            double p = (a + b + c) / 2;     //半周长
            double s = Math.Sqrt(p * (p - a) * (p - b) * (p - c));  // 海伦公式求面积
            distance = 2 * s / a;// 返回点到线的距离（利用三角形面积公式求高）

            return distance;
        }
    }
}
